//--------------------------------------------------------------------
// Atom
//--------------------------------------------------------------------
// Superclass for all particles.
//
// Usage:
//   var swarm = AtomSwarm.new(0, 2, [800, 600]).play;
//   var atom = Atom(swarm, [xpos, ypos], [xvec, yvec]);
//   swarm.add(atom);
//
//   atom.flash;
//   atom.destroy;
//
//--------------------------------------------------------------------


Atom {
	var <>env,			// container swarm
	    <>pos,			// position
	    <>vec,			// vector
	    <>rule_params,
	    <>color,
	    <>diameter = 5;
	
	var <>paused = false;
	
	var <>trail,
	    <>trailpos = 0,
	    <>trailcounter = 0,
	    <>traillength = 200,
	    <>trailthresh = 1;
	
	var <>rules,
	    <>activeRules;
	    
	var <velocity = 0,
	    <bearing = 0,
	    <>maxvelocity = 25.0;	// clamp
	
	// Wrap movement beyond environment bounds?
	// Useful, though causes semantic problems for some rules
	// (can suddenly shift environment norms)
	var wrap = true;
	
	// Used to recalculate vectors every N frames,
	// where N = 1/(env.rate)
	var counter = 0;
	
	// visual flash
	var <flashing = false,
	    <>flashwidth = 51,
	    <>flashcolor,
	    <>flashmode = 0;

	// is atom in the process of being removed?
	var <destroying = false;

	classvar <>globalRules;
	
	var <basenote = 50;
	// chance of using a note in the major scale
	var <>basenoteprob = 0.4;
	var <>basenotemutate = 0.01;
	var <>basenoteinfect = 0.4;
	var <>ringcolor,
	    <>ringdistance = 1;
	

	//----------------------------------------------------------------
	// CLASS METHODS
	//
	
	*initClass {
	
		globalRules = Dictionary[
		
			//------------------------------------------------------
			// COHESION:
			//   Attraction to the global centre of the swarm
			//   (from Reynolds)
			//------------------------------------------------------
			
			\cohesion
			-> { |self, env, params|
				var vec;
				vec = env.meanPos - self.pos;
				vec = vec  / 80.0;
				vec;
			},


			//------------------------------------------------------
			// GRAVITY:
			//   Global gravitational attraction, proportional
			//   to the inverse square distance
			//------------------------------------------------------
						
			\gravity
			-> { |self, env, params|
				var vec = 0;
				var attraction;
				
				env.atoms.select(_ != self).do
				{ |atom|
					attraction = self.distanceFrom(atom).squared.reciprocal;
					vec = vec + (50.0 * attraction * (atom.pos - self.pos))
				};
				
				vec;
			},


			//------------------------------------------------------
			// GRAVITY_BOUNDED:
			//   Gravitational attraction within defined range
			//------------------------------------------------------

			\gravity_bounded
			-> { |self, env, params|
				var vec = 0;
				var distance;
				
				env.atoms.select(_ != self).do
				{ |atom|
					distance = self.distanceFrom(atom);
					if (distance < 200)
					   { vec = vec + (50.0 * distance.squared.reciprocal * (atom.pos - self.pos)) };
				};
				
				vec;
			},


			//------------------------------------------------------
			// ATTRACTORS:
			//   Global cohesion to attractors
			//------------------------------------------------------

			\attractors
			-> { |self, env, params|
				env.attractors.do
				{ |attr|
				};
			},


			//------------------------------------------------------
			// BROWNIAN:
			//   Random motion
			//------------------------------------------------------
		
			\brownian
			-> { |self, env, params|
				var vec = 0;
				
				var freq = params[\freq] ?? 0.05;
				if (1.0.rand < freq)
				   { vec = Array.fill(env.dim, { 10.0.rand2 }) };
				
				vec;
			},


			//------------------------------------------------------
			// ALIGNMENT:
			//   Globally align vector to that of swarm
			//   (from Reynolds)
			//------------------------------------------------------

			\alignment
			-> { |self, env, params|
				var vec;
				vec = env.meanVec - self.vec;
				vec = vec / 8.0;
				vec;
			},


			//------------------------------------------------------
			// SEPARATION:
			//   avoid collision 
			//   (from Reynolds)
			//------------------------------------------------------

			\separation
			-> { |self, env, params|
				var vec = 0;
				
				var range = params[\range] ?? 20;

				env.atoms.select(_ != self).do
				{ |atom|
					
					if (self.distanceFrom(atom) < range)
					{
						vec = vec - (atom.pos - self.pos);
						
						if (self.basenoteinfect.coin)
						{
							self.basenote = atom.basenote;
							self.flash(100, Color.green);
						};

					};
					
				};
				
				vec;
			},


			//------------------------------------------------------
			// SEP_PROP:
			//   avoid collision with a force inversely
			//   proportional to the distance from peer
			//------------------------------------------------------

			\sep_prop
			-> { |self, env, params|
				var vec = 0;
				
				var range = params[\range] ?? 50;

				env.atoms.select(_ != self).do
				{ |atom|
					
					var distance = self.distanceFrom(atom);
					if (distance < range)
					{
					
						vec = vec - ((1 - (distance / range)).cubed * (atom.pos - self.pos));
					};
				};
				
				vec;
			},
			

			//------------------------------------------------------
			// SEP_DIST:
			//   Attempt to establish a fixed distance from peers
			//------------------------------------------------------

			\sep_dist
			-> { |self, env, params|
				var vec = 0;
				
				var range = params[\range] ?? 100;

				env.atoms.select(_ != self).do
				{ |atom|
					
					var distance = self.distanceFrom(atom);
					vec = vec - ((1 - (distance / range)).cubed * (atom.pos - self.pos));
				};
				
				vec;
			},


			//------------------------------------------------------
			// CENTRE:
			//   Gravitate towards the centre of the environment
			//------------------------------------------------------
			
			\centre
			-> { |self, env, params|
				var vec = (env.size / 2.0) - self.pos;
				(vec / 300);
			},
			
			
			//------------------------------------------------------
			// BOUND:
			//   Ensure atoms remain within the environment's frame
			//------------------------------------------------------

			\bound
			-> { |self, env, params|
				var vec = Array.fill(env.dim, 0); 
				self.pos.do
				{ |x, i|
					if (x < 20) { vec[i] = 10 };
					if (x > (env.size[i] - 20)) { vec[i] = -10 };
				};
				vec;
			},
			
			//------------------------------------------------------
			// CELL:
			//   Attraction to surrounding pheromone cells
			//------------------------------------------------------
			
			\cell
			-> { |self, env, params|
				var cell = floor(self.pos / (env.size / env.cellcount));
				var vec = Array.fill(env.dim, 0);
				
//				postln("own cell is " ++ cell);

				for (0 - env.cellrange, env.cellrange) { |xoff|
					var x = (cell[0] - xoff).wrap(0, env.cellcount - 1);

					for (0 - env.cellrange, env.cellrange) { |yoff|
						var y = (cell[1] - yoff).wrap(0, env.cellcount - 1);
						
						var ph = env.cells[y][x];
							
//						[ xoff, yoff, "-", ph ].postln;
						if (xoff != 0)
						   { vec[0] = vec[0] - (1 * ph * env.cellrange / xoff) };
						if (yoff != 0)
						   { vec[1] = vec[1] - (1 * ph * env.cellrange / yoff) };
					}
				};
				
				vec;
			},
		];
	}

	//------------------------------------------------------
	// Create new atom
	//------------------------------------------------------

	*new { |env, pos, vec, rule_params|
		^this.newNoInit(env, pos, vec, rule_params).init;
	}
	
	*newNoInit { |env, pos, vec, rule_params|
		var dim = env.dim;
		
		pos = pos ?? env.size.rand;
		vec = vec ?? Array.fill(dim, 0);
		if (rule_params.isNil)
		   { rule_params = Dictionary[]; };
		//postln("creating new, pos " ++ pos ++ ", vec " ++ vec ++ ", rules " ++ rule_params);
		
		^super.newCopyArgs(env, pos, vec, rule_params);
	}


	//------------------------------------------------------
	// addRule(name -> { |self, env, params| ... })
	//
	// Add an atom-specific movement rule and activate it.
	//------------------------------------------------------
	
	addRule { |rule|
		rules = rules.add(rule);
		activeRules = activeRules.add(rule.key);
		rule_params[rule.key] = Dictionary[ \weight -> 1.0 ];
		
		if (env.rule_params.includesKey(rule.key).not)
		   { env.rule_params[rule.key] = Dictionary[ \weight -> 1.0 ]; };
	}


	//------------------------------------------------------
	// Initialise atom
	//------------------------------------------------------
	
	init {
		activeRules = env.activeRules.copy;
		rules = env.rules.copy;

		Atom.globalRules.keysDo({ |ruleName|
//			postln("doing rule " + ruleName);
			if (rule_params.includesKey(ruleName).not)
			   { rule_params[ruleName] = Dictionary[] };
			if (rule_params[ruleName].includesKey(\weight).not)
			   { rule_params[ruleName][\weight] = 1.0 };
		});
		
//		postln("rules " + rules + ", active " + activeRules + ", params" + rule_params);

		// maintain trail of previous points
		trail = Array.newClear(traillength);
		trail[0] = pos.copy;
		trailthresh = trailthresh.rand;

		color = env.color;
		this.basenote = 13.rand + 45;
		this.flash;
		
		^this;
	}


	//------------------------------------------------------
	// Destroy atom
	//------------------------------------------------------

	destroy {
		this.flash(100, Color.red);
		destroying = true;
	}

	//------------------------------------------------------
	// Calculate distance from another atom
	//------------------------------------------------------
	
	distanceFrom { |other|
		// this is incorrect but sqrt introduces a great deal of overhead
		//^(abs(this.pos[0] - other.pos[0]) + abs(this.pos[1] - other.pos[1]));
		
		// same thing formulated to work in n dimensions
		//var distance = 0;
		//this.pos.size.do({ |n|
		//	distance = distance + abs(this.pos[n] - other.pos[n]);
		//});
		//
		//^distance
		
//		if (true) { ^sqrt((this.pos - other.pos).squared.sum); };
		
		// now cache distances in our environment - saves on sqrts
		^env.distances[this][other] ? inf;
	}


	//------------------------------------------------------
	// Execute rules and calculate a new motion vector.
	//------------------------------------------------------
	
	calculateVector {
		//------------------------------------------------------
		// Slow down motion according to inertia value
		//------------------------------------------------------
		vec = vec * env.inertia;

		activeRules.do({ |ruleName|
			var fn, value, weight;
			
			fn = rules[ruleName];
			value = fn.value(this, env, env.rule_params[ruleName]);
			weight = rule_params[ruleName][\weight] * env.rule_params[ruleName][\weight];

			vec = vec + (value * weight);
		});
				
		//------------------------------------------------------
		// Clamp to maximum velocity if necessary.
		//------------------------------------------------------
		velocity = vec.squared.sum.sqrt;
		if (velocity > maxvelocity)
		{
			vec = vec * (maxvelocity / velocity);
			velocity = maxvelocity;
		};
		
		bearing = atan2(vec[0], vec[1]).clip2(pi) / pi;
	}
	
	//------------------------------------------------------
	// Translate a vector into panning arguments suitable
	// for the current output channel configuration.
	//------------------------------------------------------
	
	vecToPanArgs { |vector, params|
	
		if (AtomSwarm.ambisonic)
		{
			// ambisonic panning
			^this.vecToPanAmbi(vector, params);
		} {
			if (AtomSwarm.channels == 2)
			{
				// stereo panning
				^this.vecToPanStereo(vector, params);
			} {
				// azimuth panning
				^this.vecToPanAz(vector, params);
			};
		};
	}

	vecToPanStereo { |vector, params|
		var pan, panrange;

		params = params ? Dictionary[];
		panrange = params[\panrange] ? AtomSwarm.panrange;
		pan = vector[0] / panrange;
		
		^[ \pan, pan ];
	}
	
	vecToPanAz { |vector, params|
		var panaz;
		
		// Invert because SC's Y co-ordinate begins from the top,
		// so an element beneath another has a greater Y-position.
		// We want positive Y-values to indicate a northerly direction.
		vector[1] = 0 - vector[1];
		
		panaz = atan2(vector[0], vector[1]) / pi;
				
		^[ \pan, panaz ];
	}
	
	vecToPanAmbi { |vector, params|
		var azx, azy, rho;
		
		// See above
		vector[1] = 0 - vector[1];
		

		azx = atan2(vector[0], vector[1]);
		azy = if (azy.isNil) { 0 } { atan2(vector[2], vector[0]) };
		rho = sqrt(vector.squared.sum) / 100;
//		postln("vector: " ++ vector ++ ", azx " ++ azx);		
//		postln("ambi: x " ++ azx ++ ", y " ++ azy ++ ", rho " ++ rho);
		
		^[ \azx, azx, \azy, azy, \rho, rho ];
	}

	//------------------------------------------------------
	// Called each frame to update atom's position.
	//------------------------------------------------------

	move {
		if (paused) { ^nil };
		
		if (counter >= env.rate.reciprocal)
		{
			this.calculateVector;
			counter = 0;
		};
		counter = counter + 1;
				
		pos = pos + (vec * env.rate);
		if (wrap)
		{
			pos.do({ |value, n|
				if (value < 0)
				   { pos[n] = env.size[n] + pos[n]; postln("wrapped to " ++ pos[n]) }
				   {
				     if (value > env.size[n])
				        { pos[n] = pos[n] - env.size[n]; postln("wrapped to " ++ pos[n]) };
				   };
			});
		};

		trailcounter = trailcounter + 1;
		if (trailcounter > trailthresh)
		{
			this.addToTrail;
			trailcounter = 0;
		};
	}
	
	addToTrail {
		trailpos = trailpos + 1;
		if (trailpos >= traillength)
		   { trailpos = 0 };
		   
		trail[trailpos] = pos.copy;
	}
	
	scatter { |amp = 1.0|
		vec = vec + Array.fill(env.dim, { amp * 100.rand2 });
		if (0.3.coin)
		   { basenote = basenote + 5.rand2 };

	}
	
	basenote_ { |value|
		basenote = value;
		ringcolor = Color.hsv((value % 12) / 12.0, 1, 0.5);
	}


	//------------------------------------------------------
	// Draw visual representation of atom.
	//------------------------------------------------------
	
	draw {
		var drawcolor = color.copy;
		
		var diam = 2 + (diameter * 2 * pos[2] / env.size[2]);
		var radius = diam / 2;

		//------------------------------------------------------
		// Working in 3 dimensions? Fade the atom's alpha value
		// according to its distance in the Z dimension.
		// No equivalent representation exists for dimension 4+.
		//------------------------------------------------------
		if (env.dim > 2)
		   { drawcolor.alpha = 0.4 + (0.6 * pos[2] / env.size[2]); }
		   { drawcolor.alpha = 0.95 };
		
		
		//------------------------------------------------------
		// Draw agent.
		//------------------------------------------------------
		
		drawcolor.set;
		Pen.beginPath;
		Pen.fillOval(Rect(pos[0] - radius, pos[1] - radius, diam, diam));



		//------------------------------------------------------
		// Draw trail path.
		//------------------------------------------------------
	
		Pen.beginPath;
		Pen.moveTo(Point(trail[trailpos][0], trail[trailpos][1]));
		
		// better for z-axis
		drawcolor.alpha = 0.5;
		drawcolor.set;
		
		(traillength - 2).do({ |n|
			var m = (trailpos - n - 1).wrap(0, traillength - 1);
			if (trail[m].notNil)
			{
				Pen.lineTo(Point(trail[m][0], trail[m][1]));

				// the below four lines add a z-axis to trails, but perhaps
				// add heavy processing overhead
//				Pen.width = 1.5 * trail[m][2] / env.size[2];
//				Pen.fill;
//				Pen.beginPath;
//				Pen.moveTo(Point(trail[m][0], trail[m][1]));

			};
		});
		
		drawcolor.alpha = drawcolor.alpha * 0.4;
		drawcolor.set;
		Pen.stroke;		
//		Pen.fill;

		//------------------------------------------------------
		// Draw inner ring: represents base note for melodic agents.
		//------------------------------------------------------

		Pen.beginPath;
		ringcolor.set;
		Pen.strokeOval(Rect(
			pos[0] - radius - ringdistance - 1,
			pos[1] - radius - ringdistance - 1,
			diam + 2 + (ringdistance * 2),
			diam + 2 + (ringdistance * 2)
		));

		//------------------------------------------------------
		// Create outer flash for one frame only.
		//------------------------------------------------------
		if (flashing)
		{
			// randomise flash colors
			var flashradius = floor(flashwidth / 2);
			
			drawcolor = flashcolor.copy;
			drawcolor.alpha = drawcolor.alpha * 0.3;
			drawcolor.set;
			
			switch (flashmode)
				{0}
				{
					Pen.fillOval(Rect(pos[0] - flashradius,
			                  pos[1] - flashradius,
			                  flashwidth,
			                  flashwidth));
				}
				{1}
				{
					5.do({ |n|
						var diam = flashwidth / (n + 1);
						drawcolor.alpha = drawcolor.alpha * (0.5.rand2 + 1.0);
						drawcolor.set;
						Pen.strokeOval(Rect(pos[0] - (diam / 2).floor,
			                  pos[1] - (diam / 2).floor,
			                  diam,
			                  diam));
			         });
				};
			
			flashing = false;
		};
		
		
		if (destroying)
		{
			env.atoms.remove(this);
		};
	}
	
	//------------------------------------------------------
	// Create circular flash around agent for one frame.
	//------------------------------------------------------

	flash { |width = 50, color, mode = 0|
		flashwidth = width;
		flashing = true;
		flashcolor = color ? this.color;
		flashmode = mode;
		
		this.addToTrail;
	}
	
	
	nearestAtom {
		^env.distances[this].keys.select({ |a| env.distances[this][a] != 0 }).minItem({ |a| env.distances[this][a] });
	}
}





//--------------------------------------------------------------------
// AtomMIDI
//--------------------------------------------------------------------
// Abstract superclass implementing MIDI initialisation code.
//--------------------------------------------------------------------


AtomMIDI : Atom {
	var <>mididest = 1,		// MIDI destination index
	    <>midichannel = 0,		// Channel number (from 0)
	    <>midiout;			// MIDIOut object
	    
	// Arrays (containing N objects when environment has N dimensions)
	// containing data on mapping parameters from position and velocity.
	var <>map_axis,
	    <>map_vel;
	
	init {
		super.init;
		
		this.color = Color.green;
		
		MIDIClient.init;
		
		// maps all axes to send a control value of -1
		//  - indicates that no messaging should be done.
		map_axis = Array.fill(env.dim, 0);
		env.dim.do({ |n|
			//map_axis[n] = -1;
			map_axis[n] = n;
		});
		map_vel = -1;
		
		midiout = MIDIClient.destinations.select({ |x| x.name == "VIn" }).at(0).uid;
		midiout = MIDIOut(0, midiout);
		
		^this;
	}

	play {
		
	}

	move {
		var value;
		
		super.move(env);

		map_axis.do({ |param, n|
			if ((param > -1) && (n < env.dim))
			{
				value = 127.0 * (pos[n] / env.size[n]);
				
				midiout.control(midichannel, param, value);
			};
		});
	}

	set { |key, value|
	}
}



//--------------------------------------------------------------------
// AtomMIDIChimes
//--------------------------------------------------------------------
// Implements a virtual windchime framework with MIDI mappings.
// When an Atom of this class crosses evenly-spaced 'chime' bars
// across the x-axis of the environment, a MIDI noteOn event is
// sent with a pitch proportional to its position.
//--------------------------------------------------------------------

AtomMIDIChimes : AtomMIDI {
	var <>chimecount = 12;
	var <chimebarwidth;
	var <>midibasenote = 12;
	
	init {
		super.init;
		
		// Calculate width in pixels of each chime bar
		chimebarwidth = env.size[0] / (chimecount + 1.0);
	}

	play {
		
	}

	move {
		var chimebarold, chimebarnew, chimehit, chimevel;
		
		chimebarold = (pos[0] / chimebarwidth).floor;
		super.move;
		chimebarnew = (pos[0] / chimebarwidth).floor;
		
		//----------------------------------------------------------
		// Have we passed into a new chime bar? If so, calculate
		// the index of the chime we've hit, send an appropriate
		// MIDI event, and visually flash.
		//----------------------------------------------------------
		if (chimebarold != chimebarnew)
		{
			chimehit = min(chimebarold, chimebarnew);
			chimevel = floor((velocity / maxvelocity) * 127.0);
			midiout.noteOn(midichannel, midibasenote + (chimehit * (2 + 2.rand)), chimevel);
			this.flash;
		};
	}
}




//--------------------------------------------------------------------
// AtomSynth
//--------------------------------------------------------------------
// Superclass containing base code for Atoms which implement sound
// generation using SC Synth objects.
//--------------------------------------------------------------------

AtomSynth : Atom {
	var <>synthname,
	    <>synth;
	
	var <outbus = 0;
	
	var <playing = false;
	var <>map_axis,
	    <>map_vel;
	
	classvar <>paramMap;
	
	*initClass { |classesInited|
		//------------------------------------------------------
		// Force the Warp class to be initialised before us.
		// Necessary to create the dictionaries required before
		// asSpec can be used.
		//------------------------------------------------------
		Class.initClassTree(Spec, classesInited);


		// Array of ControlSpecs used by various synth params
		paramMap = Dictionary [
			\freq		-> [ \freq.asSpec ],
			\freq_b		-> [ \freq.asSpec ],
			\freq_c		-> [ \freq.asSpec ],
			\amp			-> [ \unipolar.asSpec ],
			\amp_b		-> [ \amp.asSpec ],
			\pan			-> [ \pan.asSpec ],
			\pan_b		-> [ \pan.asSpec ],
			\delaytime	-> [ ControlSpec(0.05, 1.0, \linear) ],
			\noise		-> [ \unipolar.asSpec ],
			\crackleparam	-> [ ControlSpec(1.0, 2.0, \linear) ],
			\randrate		-> [ \unipolar.asSpec ],
			\samplerate	-> [ \freq.asSpec ],
			\uniformity	-> [ \unipolar.asSpec ],
			\phase		-> [ \phase.asSpec ],
		];
	}

	*new { |vsynthname, env, pos, vec, rule_params|
		^super.newNoInit(env, pos, vec, rule_params).synthname_(vsynthname).init;
	}

	init {
		super.init;
		
		if ((synthname.isNil) && (env.respondsTo(\synthname)))
		   { synthname = env.synthname; };
		
		map_axis = Array.fill(env.dim, nil);
		map_vel = nil;
		
		if (env.respondsTo(\map_axis))
		{
			env.dim.do({ |n|
				map_axis[n] = env.map_axis[n];
			});
			map_vel = env.map_vel;
		};

		
		// don't necessarily want to play at init time
		// this.play;
		
		//map_axis = map_axis.scramble;
	}
	
	destroy {
		synth.set(\gate, 0);
		// synths must always be gated
		super.destroy;
	}

	play { |args|
		if (playing.not)
		{
			synth = Synth(synthname, args);
			postln("creating synth, args " ++ args);
			playing = true;
		};
	}
	
	stop {
		if (playing)
		{
			synth.free;
			playing = false;
		};
	}
	
	suffix { |inChannels|
		var suffix;
		
		if (inChannels.isNil)
		   { inChannels = '' };
		
		if (AtomSwarm.ambisonic)
		{
			suffix = '_' ++ inChannels ++ 'am';// ++ AtomSwarm.channels;
		} {
			suffix = '_' ++ inChannels ++ 'p' ++ AtomSwarm.channels;
		};
		
		^suffix;
	}

	move {
		var paramSpec,
		    paramName,
		    paramValue;
		
		super.move(env);

		map_axis.do({ |param, n|
			if ((param.notNil) && (n < env.dim))
			{
				if (param.isKindOf(ArrayedCollection))
				{
					paramName = param[0];
					paramSpec = param[1];
				} {
					paramName = param;
					paramSpec = paramMap[param][0];
				};
					
				paramValue = paramSpec.map(pos[n] / env.size[n]);
				synth.set(paramName, paramValue);
			};
		});
		if (map_vel.notNil)
		{
			if (map_vel.isKindOf(ArrayedCollection))
			{
				paramName = map_vel[0];
				paramSpec = map_vel[1];
			} {
				paramName = map_vel;
				paramSpec = paramMap[map_vel][0];
			};
			
			paramValue = paramSpec.map(velocity / maxvelocity);

			synth.set(paramName, paramValue);
		};
	}
	
	set { |key, value|
		//postln("set " ++ synth ++ " key " ++ key ++ " to " ++ value);
		synth.set(key.asSymbol, value);
	}
	
	outbus_ { |value|
		outbus = value;
		synth.set(\outbus, value);
	}
}

AtomPanner : AtomSynth
{
	var <panner_synth,
	    <panner_bus;
	
	init {
		super.init;
		
		panner_bus = Bus.audio(Server.default, 1);
		postln("spawning " ++ 'atompanner' ++ this.suffix(1));
		panner_synth = Synth('atompanner' ++ this.suffix(1), [ \inbus, panner_bus.index, \outbus, outbus ]);
		postln("creating panner, args " ++ [ \inbus, panner_bus.index, \outbus, outbus ]);
	}
	
	destroy {
		panner_bus.free;
		super.destroy;
	}
	
	play { |args|
		super.play(args ++ [ \outbus, panner_bus.index ]);
	}
	
	outbus_ { |value|
		outbus = value;
		panner_synth.set(\outbus, value);
	}
}

AtomPannerXY : AtomPanner
{
	move {
		var panargs = this.vecToPanArgs(pos - (env.size / 2.0),  Dictionary[ \panrange -> env.size[0] ]);
		if (0.002.coin)
		{
			pos = env.size.rand;
			Synth(\inst_teleport, [ \outbus, outbus ]);
		};
		
		super.move;
		
		panner_synth.set(*panargs);
	}
}



//--------------------------------------------------------------------
// AtomBeetle
//--------------------------------------------------------------------

AtomBeetle : AtomPannerXY
{
	init {
		super.init;
		color = Color.black;
		rules[\separation] = 0;
		map_vel = [ \freq, ControlSpec(2, 20) ];
	}
	
	move {
		super.move;
		env.atoms.select({ |a| a != this }).do({ |atom|
			var dist = atom.distanceFrom(this);
			if (dist < 100)
			{
				vec.size.do({ |n|
					atom.vec[n] = atom.vec[n] - (0.5.rand * (this.vec[n] - atom.vec[n]));
				});
			};
		});
	}
}



//--------------------------------------------------------------------
// AtomSynthChimes
//--------------------------------------------------------------------
// As AtomMIDIChimes, but creates a synth object with mapped frequency.
//--------------------------------------------------------------------

AtomSynthChimes : AtomPanner {
	var <chimecount = 12;
	var <chimebarwidth;
	
	*initClass {

	}
	
	init {
		super.init;
		synthname = 'sonar';
		this.chimecount_(chimecount);
	}
	
	chimecount_ { |v|
		chimecount = v;
		chimebarwidth = env.size[0] / (chimecount + 1.0); 
	}

	move {
		var chimebarold, chimebarnew, chimehit;
		
		chimebarold = (pos[0] / chimebarwidth).floor;
		super.move;
		chimebarnew = (pos[0] / chimebarwidth).floor;
		if (chimebarold != chimebarnew)
		{
			chimehit = min(chimebarold, chimebarnew);
			synth = Synth(synthname, [
				\outbus, 	panner_bus.index,
				\amp, 	velocity / maxvelocity,
				\freq, 	midicps(basenote + 12 + degreeToKey(chimehit + 1, [ 0, 5, 2, 7 ])),
			 ]);
			synth.set(\gate, 0);
			this.flash;
			//postln("hit chime " ++ chimehit);
		};
	}
}





//--------------------------------------------------------------------
// AtomGranulator
//--------------------------------------------------------------------
// Continually sends its synth object a panaz parameter corresponding
// to its rotational bearing from the centre of its environment.
//--------------------------------------------------------------------

AtomGranulator : AtomPanner {
	var <>buffer, <>sample;
	
	*initClass {
		paramMap.putAll(
		Dictionary[
			\density	-> [ ControlSpec(20, 500, \exponential) ],
			\rate	-> [ ControlSpec(0.1, 3, \exponential) ],
			\pos		-> [ \unipolar.asSpec ],
			\dur		-> [ ControlSpec(0.04, 0.4, \exponential) ],
		]);


		SynthDef(\atomgranulator_1p2, { |outbus = 0, bufnum = 0,
								 pos = 0.0, density = 100.0, rate = 1.0, dur = 0.1,
								 amp = 0.5, pan = 0.0, interp = 4, gate = 1|
			var data, clock, env;
		//	clock = Impulse.ar(density + WhiteNoise.ar(50));
			clock = Dust.ar(density);
			data = TGrains.ar(2, clock, bufnum, rate, pos * BufDur.ir(bufnum), dur, pan, amp * 0.3, interp);
			env = EnvGen.kr(Env.asr(10.0, 1.0, 3.0), gate: gate, doneAction: 2);
			Out.ar(outbus, data * env);
		});
		
		[2,4,6,8].do({ |c|
			SynthDef('atomgranulator_1az' ++ c, { |outbus = 0, bufnum = 0,
									 pos = 0.0, density = 100.0, rate = 1.0, dur = 0.1,
									 amp = 0.5, panaz = 0.0, panwidth = 2.0, interp = 4, gate = 1|
				var data, clock, env;
				clock = Dust.ar(density);
				data = TGrains.ar(c, clock, bufnum, rate, pos * BufDur.ir(bufnum), dur, panaz, amp * 0.3, interp);
			//	data = PanAz.ar(4, Mix.ar(data), panaz, width: panwidth);
				env = EnvGen.kr(Env.asr(10.0, 1.0, 3.0), gate: gate, doneAction: 2);
				Out.ar(outbus, data * env);
			});
		});
	}
	
	*new { |vsample, env, pos, vec, rule_params|
		^super.newNoInit(env, pos, vec, rule_params).sample_(vsample).init;
	}
	
	play {
		synth = Synth(synthname, [
			\outbus, panner_bus.index, 
			\bufnum, -1,
			\density, 100, \rate, 1.0,
			\pos, 0,
			\pan, 0 // 1.0.rand2
		]);
	}
	
	init {		
		super.init;

		if (AtomSwarm.channels == 2)
		   { synthname = 'atomgranulator_1p2' }
		   { synthname = 'atomgranulator_1az' ++ AtomSwarm.channels };
		
		map_axis = [ \pos, \dur, \density ];
		map_vel = \amp;
		
		color = Color(0.7, 0.7, 0.8);
		
		if (sample.isNil)
		   { Error("No soundfile specificed for granulation").throw };
				
		this.play;

		buffer = Buffer.read(Server.default, sample, action: {
			synth.set(\bufnum, buffer.bufnum);
		});
	}
	
	move {
		var panargs = this.vecToPanArgs(pos - (env.size / 2.0),  Dictionary[ \panrange -> env.size[0] ]);

		super.move;
		
		panner_synth.set(*panargs);
	}

	destroy {
		synth.set(\gate, 0);
//		buffer.free;
		super.destroy;
	}
}




//--------------------------------------------------------------------
// AtomPlayer
//--------------------------------------------------------------------
// Plays a looped soundfile, with amplitude and pan relative to the
// (single) AtomListener object in the same environment.
//--------------------------------------------------------------------

AtomPlayer : AtomPanner {
	var <>sample, <>volume, <>attack, <>release;
	var buffer, <>synth, <>synthname;
	
	var <>listener_dist = 0;
	
	*initClass
	{

		//-----------------------------------------------------------
		// SynthDefs:
		//   - atomplayer_N plays an N-channel buffer with gated
		//     env, with stereo panning
		//   - atomplayer_Naz plays an N-channel sound buffer with
		//     gated env, with PanAz rotational azimuth panning
		//-----------------------------------------------------------


		SynthDef.writeOnce(\atomplayer_1, { |outbus = 0, bufnum = 0, amp = 0.0, pan = 0.0, rate = 1.0,
		                                   volume = 0.5, attack = 1.0, gate = 1, release = 1.0|
			var out, env;
			out = PlayBuf.ar(1, bufnum, rate * BufRateScale.ir(bufnum), loop: 1);
			env = EnvGen.kr(Env.asr(attack, 1, release), gate: gate, doneAction: 2);
			out = out * env * amp * volume;
			Out.ar(outbus, Pan2.ar(out, pan));
		});
		
		SynthDef.writeOnce(\atomplayer_2p2, { |outbus = 0, bufnum = 0, amp = 0.0, pan = 0.0, rate = 1.0,
		                                   volume = 0.5, attack = 1.0, gate = 1, release = 1.0|
			var out, env;
			out = PlayBuf.ar(2, bufnum, loop: 1);
			env = EnvGen.kr(Env.asr(attack, 1, release), gate: gate, doneAction: 2);
			out = out * env * amp * volume;
			Out.ar(outbus, Pan2.ar(Mix.ar(out), pan));
		});
		
		[2,4,6,8].do({ |c|
			SynthDef.writeOnce(\atomplayer_1az ++ c, { |outbus = 0, bufnum = 0, amp = 0.0, panaz = 0.0, panwidth = 2.0,
			                                   volume = 0.5, attack = 1.0, gate = 1, release = 1.0|
				var out, env;
				out = PlayBuf.ar(1, bufnum, loop: 1);
				env = EnvGen.kr(Env.asr(attack, 1, release), gate: gate, doneAction: 2);
				out = out * env * amp * volume;
				Out.ar(outbus, PanAz.ar(c, out, panaz, width: panwidth));
			});
			
			SynthDef.writeOnce(\atomplayer_2az ++ c, { |outbus = 0, bufnum = 0, amp = 0.0, panaz = 0.0, panwidth = 2.0,		                                   volume = 0.5, attack = 1.0, gate = 1, release = 1.0|
				var out, env;
				out = PlayBuf.ar(2, bufnum, loop: 1);
				env = EnvGen.kr(Env.asr(attack, 1, release), gate: gate, doneAction: 2);
				out = out * env * amp * volume;
				Out.ar(outbus, PanAz.ar(c, Mix.ar(out), panaz, width: panwidth));
			});
		});
	}
	
	*new { |vsample, env, pos, vec, rule_params|
		^super.newNoInit(env, pos, vec, rule_params).sample_(vsample);
	}
	
	init {
		super.init;
		
		volume = 0.5;
		attack = 1.0;
		release = 1.0;

		// Want to ignore normal axis mappings - these will
		// be controlled by a corresponding Listener object
		map_axis = Array.fill(env.dim, nil);
		map_vel = nil;
		
		color = Color(0.8, 0.4, 0.4);
		
		buffer = Buffer.read(Server.default, sample, action: {
		
			synthname = 'atomplayer_1';
//			if (AtomSwarm.channels == 2)
//			   { synthname = 'atomplayer_' ++ buffer.numChannels ++ 'p2' }
//			   { synthname = 'atomplayer_' ++ buffer.numChannels ++ 'az' ++ AtomSwarm.channels };
			
			postln("synthname " ++ synthname);
			//synthname.postln;
			synth = Synth(synthname, [
				\outbus, 	panner_bus.index, 
				\bufnum, 	buffer.bufnum,
				\volume, 	volume * 2.4,
				\attack, attack,
				\release, release,
				\amp, 0.1,
			]);
		});
	}
	
	move {
		var listener;
		var listener_dist_new,
		    listener_delta;
		
		var amp, panArgs;

		super.move;

		listener = env.listener;		
		listener_dist_new = this.distanceFrom(listener);
		listener_delta = listener_dist_new - listener_dist;
		listener_dist = listener_dist_new;
		
		if (listener_dist < AtomSwarm.audiorange)
		{
			amp = (AtomSwarm.audiorange - listener_dist) / AtomSwarm.audiorange;
			amp = amp.cubed;

			panArgs = this.vecToPanArgs(pos - listener.pos);
		} {
			amp = 0;
		};
		
		synth.set(\amp, amp, \rate, 1.0 - (0.05 * listener_delta).clip2(0.3));
		panner_synth.set(*panArgs);	
	}
}

AtomListener : Atom {
	init {
		super.init;
		color = Color.red;
		^this;
	}
}



//--------------------------------------------------------------------
// AtomCFollow
//--------------------------------------------------------------------
// Follows the motion of the mouse cursor using variables set
// within the AtomSet object.
//--------------------------------------------------------------------

AtomCFollow : Atom {
	var <>window;
	
	*new { |env, pos|
		var self = super.newNoInit(env, pos).window_(env.window).init;
		^self;
	}

	init {
		var winleft, wintop;
		
		super.init;

		winleft = window.bounds.left;
		wintop = AtomSet.screenHeight - window.bounds.bottom;
		color = Color(0.9, 0.5, 0.5);
		
		if (AtomSet.mouseX.isNil)
		   { Error("Server must be running for cursor tracking").throw };
		
		this.addRule(\cursor -> { |self, env|
			var vec = [AtomSet.mouseX - winleft, AtomSet.mouseY - wintop] - self.pos;
			(vec / 100.0);
		});
	}
}



//--------------------------------------------------------------------
// AtomCollider
//--------------------------------------------------------------------
// Plays back an audio buffer when its distance from any other Atom
// falls below a specified range, with amplitude proportional to
// the collision velocity and azimuth panning around the centre of
// the environment.
//--------------------------------------------------------------------

AtomCollider : AtomPanner
{
	var <>randrate = 0.3;
	var <>synthname;
	
	*initClass
	{
	
		//-----------------------------------------------------------
		// SynthDefs:
		//  - atomcollider_1pN plays a 1-channel buffer panned over
		//    stereo channels
		//  - atomcollider_1azN plays a 1-channel buffer with azimuth
		//    panning over N speakers.
		//-----------------------------------------------------------
		
		SynthDef.writeOnce('atomcollider_1p2', { |bufnum = 0, outbus = 0, rate = 1.0, amp = 0.5, pan = 0|
	        var signal;
	        signal = PlayBuf.ar(1, bufnum, rate * BufRateScale.kr(bufnum));
	        FreeSelfWhenDone.kr(signal);
	        signal = signal * amp;
	        Out.ar(outbus, Pan2.ar(signal, pan));
		});

		[2,4,6,8].do({ |c|
			SynthDef.writeOnce('atomcollider_1az' ++ c, { |bufnum = 0, outbus = 0, rate = 1.0, amp = 0.5, panaz = 0|
		        var signal;
		        signal = PlayBuf.ar(1, bufnum, rate * BufRateScale.kr(bufnum));
		        FreeSelfWhenDone.kr(signal);
		        signal = signal * amp;
		        Out.ar(outbus, PanAz.ar(c, signal, panaz));
			});
		});
		
		SynthDef.writeOnce('atomcollider_1am', { |bufnum = 0, outbus = 0, rate = 1.0, amp = 0.5, azx = 0, azy = 0, rho = 1|
	        var signal, bf;
	        
	        signal = PlayBuf.ar(1, bufnum, rate * BufRateScale.kr(bufnum));
	        FreeSelfWhenDone.kr(signal);
	        signal = signal * amp;
	        
	        bf = BFEncode1.ar(signal, azx, azy, rho);
	        bf = BFDecode1.ar(bf[0], bf[1], bf[2], bf[3],
			[-0.25pi, 0.25pi, 0.75pi, 1.25pi,-0.25pi, 0.25pi, 0.75pi, 1.25pi], 
			[0.25pi,0.25pi,0.25pi,0.25pi,-0.25pi,-0.25pi,-0.25pi,-0.25pi]
		   );
		   
		   Out.ar(outbus, bf);
		});
	}
	
	init
	{
		super.init;
		
		synthname = 'atomcollider' ++ this.suffix(1);
		synthname = 'one_cricket';
		
		//-----------------------------------------------------------
		// Overwrite separation rule. The new rule follows the same
		// form, but also plays a sound buffer corresponding to
		// the atom's index in its population.
		//-----------------------------------------------------------
		rules[\separation] = { |self, env, params|
			var vec = 0;
			var id,
			    panargs,
			    amp,
			    args;
			
			var range = params[\range] ?? 20;
	
			env.atoms.select(_ != self).do
			{ |atom|
				
				if (self.distanceFrom(atom) < range)
				{
					id = env.atoms.indexOf(self);
					amp = (velocity / maxvelocity) * 0.4;
					
					args = [
						\bufnum,	id + 2,
						\outbus,	panner_bus.index,
						\rate,	1.0 + randrate.rand2,
						\amp,	amp * 2.5,
					];
					
					panargs = atom.vecToPanArgs(atom.pos - (env.size / 2.0), Dictionary[ \panrange -> env.size[0] ]);
					panner_synth.set(*panargs);
					
					TempoClock.default.sched(0.05.rand, {
						Synth(synthname, args)
					});
					
					self.flash;
					vec = vec - (atom.pos - self.pos);
				};
			};
			
			vec;
		};
		
		^this;
	}
}




//--------------------------------------------------------------------
// AtomNearest
//--------------------------------------------------------------------
// Plays back an audio buffer when the nearest Atom changes.
//--------------------------------------------------------------------

AtomNearest : AtomPanner
{
	var <nearest_prev;
//	    <synthname;
	
	init {
		super.init;
		color = Color(0.2, 0.5, 0.9);
	}
	
	move {
		var nearest;
		
		super.move;
		nearest = this.nearestAtom;
		
		if (nearest != nearest_prev)
		{
			synthname = 'atomcollider' ++ this.suffix(1);
			
			nearest_prev = nearest;
			this.flash;
			nearest.flash;
			Synth(synthname, [
				\bufnum, 10.rand,
				\outbus, panner_bus.index, 
				\rate, 0.7 + 0.15.rand2,
				\amp, 0.1 + (0.2 * velocity / maxvelocity);
			]);
		};
	}
}


//--------------------------------------------------------------------
// AtomDepositor
//--------------------------------------------------------------------
// Creates a synth object after travelling a fixed distance in pixels.
//--------------------------------------------------------------------

AtomDepositor : AtomPanner
{
	var <distance = 0;
	var <>deposit_period = 50;
	// distance travelled, in pixels
	
	init {
		super.init;
		color = Color.blue;
	}
	
	move {
		var periods, periods_old;
		
		super.move;
		distance = distance + (velocity * env.rate);

		if (distance > deposit_period)
		{		
			this.trigger;
			distance = distance - deposit_period;
		};
	}
	
	trigger {
			var nearestAtom = this.nearestAtom;
			var damp = 0.95;
			var panArgs = [ \pan, 1.0.rand2 ];
			var scale = [ 0, 2, 3, 5, 7, 8, 11 ];
			var freq;
			if (basenoteprob.coin) { scale = [ 0, 2, 7 ] };
			if (basenotemutate.coin) {
				this.basenote = basenote + 10.rand2;
				this.flash(Color.yellow);
			};

			freq = midicps(velocity.round.degreeToKey(scale) + basenote);
//			postln("base note " ++ basenote ++ ", freq " ++ freq);

			
			if (nearestAtom.notNil)
			{
				var nearestDist = env.distances[this][nearestAtom];
				damp = (nearestDist / 400).clip(0, 0.4).squared;
				damp = 1.0 - damp;
				
				panArgs = this.vecToPanArgs(pos - nearestAtom.pos);
			};

			this.flash(damp * 300 - 230);
			damp = damp.clip(0.1, 0.997);
			
			Synth(\inst_pluck, [ \outbus, panner_bus.index, \freq, freq, \damp, damp ]);
			panner_synth.set(*panArgs);

			// send to 0 to bypass reverb
//			Synth(\gli_1, [ \outbus, 0, \freq, freq, \damp, damp ] ++ panArgs);
//			Synth(\inst_bell, [ \outbus, 16, \freq, 30 + (velocity * 10) ]);
	}
}

//--------------------------------------------------------------------
// AtomDepositorGlitch
//--------------------------------------------------------------------
// AtomDepositor which generates a synth with randomly applied effects.
//--------------------------------------------------------------------

AtomDepositorGlitch : AtomDepositor
{
	init {
		// bypass fx on bus 16
		super.init;
		panner_synth.set(\outbus, 0);
		deposit_period = 200;
	}
	
	trigger {
		var fx_outbus = panner_bus.index;
		var amp = velocity / maxvelocity;
		var nearest = this.nearestAtom;
		
		if (nearest.notNil)
		{
			var panArgs = this.vecToPanArgs(pos - nearest.pos);
			panner_synth.set(*panArgs);
		};
		   
		this.flash(20 + (amp * 60));
		
		if (0.1.coin)
		{
//			fx_outbus = this.randfx;
		};

		// send to 0 to bypass reverb
//		Synth(\gli_1, [ \outbus, fx_outbus, \amp, amp ]);
		Synth(\gli_1, [ \outbus, 0, \amp, amp ]);		
	}
	
	randfx {
		var fx_name = [
			\granule,
			\pitchshift,
			\rev_lyon,
			\rev_iron,
			\delay_seaside,
			\ringmod,
			\degrader,
			\dist3,
			\invert,
		].choose;

		var fx_args;
		var fx;
		var bus;

		bus = Bus.audio(Server.default, 2);

		fx_args = [ \wet, 1.0, \inbus, bus.index, \outbus, panner_bus.index ]; 
		
		SIFx.fxMap[fx_name][1].do({ |param|
			fx_args = fx_args ++ [ param, SICore.paramMap[param][2].map(1.0.rand) ];
		});
		if (fx_name == \dist3) { fx_args = fx_args ++ [ \amp, 0.2 ] };

		postln("Creating " ++ fx_name ++ ": " ++ fx_args);

		fx = Synth('fx_' ++ fx_name, fx_args);
		TempoClock.default.sched(5, {
			postln("freeing fx: " ++ fx_name);
			fx.free;
			bus.free;
		});
		
		^bus.index;
	}
}


//--------------------------------------------------------------------
// AtomVelThresh
//--------------------------------------------------------------------
// Triggers a synth object after reaching a given velocity.
// After dropping beneath a minimum velocity threshold, this synth
// is then silenced (gate set to 0).
//--------------------------------------------------------------------

AtomVelThresh : AtomPanner
{
	var threshold = 10;
	var gate = false;
	
	init {
		super.init;
		
		synthname = \atom_velsine;
		this.play([ \outbus, panner_bus.index, \gate, 0 ]);
	}
	
	destroy {
		synth.free;
		super.destroy;
	}
	
	move {
		super.move;
		
		if (velocity > threshold)
		{
			if (gate.not)
			{
				var scale = [ 0, 2, 3, 5, 7, 8, 11 ];
				var freq;
				if (basenoteprob.coin) { scale = [ 0, 2, 7 ] };
				if (basenotemutate.coin) {
					this.basenote = basenote + 10.rand2;
					this.flash(Color.yellow);
				};
	
				freq = midicps((15.rand).round.degreeToKey(scale) + basenote - 12);
//				synth.set(\freq, 50 * 2.pow(5.rand), \gate, 1);
				synth.set(\freq, freq);
				synth.set(\gate, 1);
					
				gate = true;
				this.flash;
				this.color = Color(1.0, 0.5, 0.0);
			};
		}
		{
			if (gate)
			{
				synth.set(\gate, 0);
				gate = false;
				this.color = Color.white;
			};
		};
		
		if (gate)
		{
			var nearestAtom = this.nearestAtom;
			var panArgs = [ \pan, 1.0.rand2 ];

			if (nearestAtom.notNil)
			{
				panArgs = this.vecToPanArgs(pos - nearestAtom.pos);
			};
	
//			postln("set " ++ panArgs);
			panner_synth.set(*panArgs);
		};
	}
}




AtomVisualWarp : Atom
{
	move {
		super.move;
	}
}


//--------------------------------------------------------------------
// AtomDoppler
//--------------------------------------------------------------------
// Generates a drone with an emulated doppler effect, whose intensity
// is based on the relative velocity of the nearest peer.
//--------------------------------------------------------------------


AtomDoppler : AtomPanner
{
	var nearest_dist = 0;
	
	init {
		super.init;
		
		synthname = 'hornet';
		map_axis = Array.fill(env.dim, nil);
		map_vel = [ \amp, ControlSpec(0.1, 0.4) ];
//		this.play([ \octave, 4.rand - 1 ]);

		this.play([ \octave, [ 1, 2, 3, 4 ].choose ]);
		
		color = Color.yellow;

		panner_synth.set(\outbus, 0);
	}
	
	move {
		var nearest,
		    nearest_dist_new,
		    nearest_delta;
		 
		var panArgs;
		
		super.move;
		
//		nearest = this.nearestAtom;
		nearest = env.listener;
		if (nearest.notNil)
		{
			nearest_dist_new = this.distanceFrom(nearest);
		} {
			nearest_dist_new = 0;
		};

		nearest_delta = nearest_dist_new - nearest_dist;
		nearest_dist = nearest_dist_new;
		
		panArgs = this.vecToPanArgs(pos - nearest.pos);
		
		synth.set(\freq, basenote.midicps + (nearest_delta));
		panner_synth.set(*panArgs);
	}

}


//--------------------------------------------------------------------
// AtomEffector
//--------------------------------------------------------------------
// 'Captures' nearby atoms based on a random function. Captured atoms
// are then passed through an effect synth, whose parameters are
// controlled by the Effector's motion.
//--------------------------------------------------------------------


AtomEffector : AtomSynth
{
	var <>nearest_prev;
	
	var <>synth_gated = false;
	
	var <>capture_chance = 0.1,
	    <>capture_thresh = 100;
	var <>release_chance = 0.001,
	    <>release_thresh = 500;

	var <>bus;
	
	init {
		super.init;
		
		// assume stereo for now
		bus = Bus.audio(Server.default, 2);
		
		ringdistance = capture_thresh.sqrt;

		switch (9.rand)
			{ 0 }
			{
				synthname = 'fx_ringmod';
				map_axis[0] = \freq;
				map_axis[1] = [ \wet, \unipolar.asSpec ];
			}
			{ 1 }
			{
				synthname = 'fx_degrader';
				map_axis[0] = [ \samplerate, 	[ 100, 10000, \linear ].asSpec ];
				map_axis[1] = [ \bits,			[ 3, 16, \linear ].asSpec ];
			}
			{ 2 }
			{
				synthname = 'fx_pitchshift';
				map_axis[0] = [ \pitchratio,		[ 0.1, 10, \exp ].asSpec ];
			}
			{ 3 }
			{
				synthname = 'fx_delay_seaside';
				map_axis[0] = [ \delay_time,		[ 0.01, 0.2, \linear ].asSpec ];
				map_axis[1] = [ \decay_time,		[ 0.1, 5.0, \linear ].asSpec ];
			}
			{ 4 }
			{
				synthname = 'fx_feedli';
				map_axis[0] = [ \feedback,			[ 0.6, 1.0, \linear ].asSpec ];
				map_axis[1] = [ \delaytime,			[ 0.01, 0.4, \linear ].asSpec ];
			}
			{ 5 }
			{
				synthname = 'fx_dist3';
				map_axis[0] = [ \drive,				\unipolar.asSpec ];
				map_vel =		[ \wet,				\unipolar.asSpec ];
				map_axis[1] = [ \amp,				[ 0.1, 0.2 ].asSpec ];
			}
			{ 6 }
			{
				synthname = 'fx_rev_iron';
				map_axis[0] = [ \reverbtime, [ 2.0, 10.0 ].asSpec ];
				map_vel = 	[ \wet, [ 1, 2 ].asSpec ];
			}
			{ 7 }
			{
				synthname = 'fx_granule';
				map_vel = [ \pitchratio,		[ 0.1, 10, \exp ].asSpec ];
				map_axis[0] = [ \windowsize,	[ 0.01, 0.2 ].asSpec ];
				map_axis[1] = [ \amp,			[ 1, 2 ].asSpec ];
			}
			{ 8 }
			{
				synthname = 'fx_charsiesis';
				synth_gated = true;
				
				map_vel = 	[ \feedback,		[ 0, 0.9 ].asSpec ];
				map_axis[0] = [ \rate_range,	[ 0.01, 5.0, \exp ].asSpec ];
				map_axis[1] = [ \delay_range,	[ 0.01, 3.0, \exp ].asSpec ];
				map_axis[2] = [ \lpf_f,			[ 300, 10000, \exp ].asSpec ];
			}
			;
//		this.play([ \wet, 1.0, \outbus, 0, \inbus, bus.index]);
		this.play([ \wet, 1.0, \outbus, outbus, \inbus, bus.index]);
		postln("fx " ++ synthname ++ ", " ++ "out" ++ outbus);

	}
	
	play { |args|
		if (playing.not)
		{
			postln("NEW " ++ synthname ++ ", " ++ args);
			synth = Synth.tail(Server.default, synthname, args);
			playing = true;
		};
	}
	
	inbus {
		^bus.index;
	}
	
	draw {
		ringcolor = Color(1, 1, 1, 0.5);

		super.draw;

//		Color(1, 1, 1, 0.1).set;
		// brighter for projector
		Color(1, 1, 1, 0.6).set;
		
		env.atoms.select({ |a| (a != this) && a.respondsTo('outbus') }).do({ |atom|

			var dist = atom.distanceFrom(this);
		
			if (atom.outbus == bus.index)
			{
				Pen.line(Point(this.pos[0], this.pos[1]), Point(atom.pos[0], atom.pos[1]));
				Pen.stroke;
				
				if ((dist > release_thresh) || (release_chance.coin))
				{
					postln("releasing - " ++ env.set.outbus);
					atom.outbus = env.set.outbus;
				};
			} {
				if ((dist < capture_thresh) && (capture_chance.coin))
				{
					postln("CAPTURING - bus " ++ bus.index);
					atom.outbus = bus.index;
					atom.flash;
				};
			};
		});
	}
	
	scatter {
		super.scatter;
		env.atoms.select({ |a| (a != this) && a.respondsTo('outbus') }).select({ |a| a.outbus == bus.index }).do({ |atom|
			if (0.3.coin)
			   { atom.outbus = env.set.outbus };
		});
	}
	
	destroy {

		if (synth_gated)
		{
			synth.set(\gate, 0);
			env.atoms.select({ |a| a.respondsTo('outbus') }).select({ |a| a.outbus == bus.index }).do({ |atom|
				atom.outbus = env.set.outbus;
			});
			bus.free;
		} {
			// assumes that wetness parameter has a Lag of 0.5s
			synth.set(\wet, 0.0);
			
			SystemClock.sched(0.6, {
				env.atoms.select({ |a| a.respondsTo('outbus') }).select({ |a| a.outbus == bus.index }).do({ |atom|
					atom.outbus = env.set.outbus;
				});
				bus.free;
				synth.free;
			});
		};
		
		super.destroy;
	}

	outbus_ { |value|
		^0;
//		synth.set(\outbus, value);
	}
}

//--------------------------------------------------------------------
// AtomNode
//--------------------------------------------------------------------
// An element in a network of nodes, with a graphical representation
// of a connection between peers.
//--------------------------------------------------------------------

AtomNode : AtomSynth
{
	var peers;
	
	init {
		super.init;
		peers = [];
	}
	
	addPeer { |atom|
		peers = peers.add(atom);
	}
	
	draw {
		super.draw;
		
		peers.do({ |atom|
			Pen.use {
				Color(1, 1, 1, 0.2).set;
				Pen.line(Point(pos[0], pos[1]), Point(atom.pos[0], atom.pos[1]));
				Pen.stroke;
			};
		});
	}
}

AtomNodeHead : AtomNode
{
	draw {
		super.draw;
		Pen.use {
			Color(1, 1, 1, 0.2).set;
			Pen.strokeOval(Rect(pos[0] - 7, pos[1] - 7, 15, 15));
		};
	}
}