//--------------------------------------------------------------------
// AtomSet
//--------------------------------------------------------------------
// Contains global config required for a performance.
//  - Installs MIDI and keyboard control bindings
//  - Spawns and polls the Synth required for cursor tracking
//  - Preloads audio buffers 
//  - Sets initial values for rule parameters
//  - Creates synths for custom effects
//--------------------------------------------------------------------




AtomSet
{
	var <>server;
	
	var <>swarms,			// Array of swarms (though only 1 can be controlled at present)
	    <>win_disp,		// Display window (AtomSwarmWindow)
	    <>win_ctrl,		// Control window (AtomCtrlWindow)
	    <>pauserate;		// Speed at which Atoms come to a halt when UI is paused
	    					// (space bar)
	
	
	var <>traillength = 0;
	
	var <>fx_bus;
	
	var <>group_aux;
	
	var <>synth_reverb,
	    <>synth_reverb_auto,
	    <>synth_amp,
	    <>synth_compress;
	
	var <>synth_fb;
	    
	var <>ambi_decoder,
	    <>ambi_bus;
	
	var <monome;
	
	var <>midi = false;

	var <>midicontrol,
	    <>midibase = 48;
	    
	// show comedic intro for new classes
	var <>introductions = true,
	    <>introductions_done;
	
	classvar <sampleRoot = "/Users/daniel/audio/samples";
	classvar <bindings,
	         <atomtypes;


	classvar <>mouseX, <>mouseY;
	classvar <screenWidth = 1440, <screenHeight = 900;
	classvar mouse_responder,
	         mouse_synth,
	         <>mouse_inited,
	         <mouseDownActions;

	classvar onset_responder,
	         onset_buffer,
	         onset_synth,
	         onset_inited;
	
	*initClass {
		atomtypes = Dictionary[
			\collider 	-> [ $b, AtomCollider, "Collider" ],
			\chimes		-> [ $c, AtomSynthChimes, "Chimes" ],
			\follower		-> [ $f, AtomCFollow, "Follower" ],
			\nearest		-> [ $n, AtomNearest, "Proxim" ],
			\depositor	-> [ $d, AtomDepositor, "Depositor" ],
			\glitch		-> [ $x, AtomDepositorGlitch, "Glitch" ],
			\velthresh	-> [ $v, AtomVelThresh, "Vel Thresh" ],
			\doppler		-> [ $o, AtomDoppler, "The Hornet" ],
			\synth		-> [ $s, AtomSynth, "Synth" ],
			\effector		-> [ $e, AtomEffector, "The Spider" ],

			\beetle		-> [ $z, AtomBeetle, "The Beetle",
			{ |env|
				var atom = AtomBeetle(\beetle, env);
//				atom.color = Color.black;
//				atom.map_vel = [ \freq, ControlSpec(2, 20) ];
				atom.outbus = 0;
				atom.play;
				env.add(atom);
			} ],

			\miditicks	-> [ $t, AtomMIDIChimes, "MIDI Chimes",
			{ |env|
				var atom;
				atom = AtomMIDIChimes(env).midichannel_(3.rand).midibasenote_(40);
				env.add(atom);
			} ],
			
			\player		-> [ $p, AtomPlayer, "The Organ",
			{ |env|
				var sample;
				
//				sample = SampleLib.glob("field/20070208 - teufelsberg/*.aif").choose;
				sample = SampleLib.glob("atmos/auscultation/1*.wav").choose;
//				[
//					"to-use/lucier_late.wav",
//					"to-use/cicadas.wav",
//					"to-use/pinknoise_2k.wav",
//					"to-use/jungle_birds.wav",
//					"to-use/swamp_crickets.wav",
//					"to-use/campfire.wav",
//					"to-use/desert_crickets.wav",
//				].choose;
				
				
				// XXX: need to get abs/rel patgh
				// env.addWithSample(AtomSet.sampleRoot ++ "/" ++ sample);
				env.addWithSample(sample);
			} ],
			
			\granulator 	-> [ $g, AtomGranulator, "The Granulator",
			{ |env|
				var atom, sample;
				
				sample =
				[
					"to-use/ligeti_apparitions_one.wav",
//					"to-use/ligeti_apparitions_three.wav",
//					"to-use/glitch_improv.wav",
//					"to-use/stocky_clap.wav",
					"to-use/cicadas.wav",
					"to-use/Stocky_laugh.wav",
					"to-use/lucier_late.wav",
//					"speech/yorkshire_tea_advert.aif",
//					"speech/trills and non-pulmonic/01.aif",
//					"speech/wigeon.aif",
				].choose;

				atom = AtomGranulator(AtomSet.sampleRoot ++ "/" ++ sample, env);
				env.add(atom);
			} ],
		];

		//-------------------------------------------------
		// Initialization code for cursor tracking.
		screenWidth = SCWindow.screenBounds.width;	
		screenHeight = SCWindow.screenBounds.height;
		
		mouse_responder = OSCresponder(nil, '/tr', { |time, rsp, msg|
			var id = msg[2];
			var pos = msg[3];
			if (id == 0)
			   //-----------------------------------------
			   // Button pressed
			   // Execute list of handler functions.
			   //-----------------------------------------
			   {
			     mouseDownActions.do({ |fn|
			     	fn.value(mouseX, mouseY);
			     });
			   }
			   {
			     if (id == 1)
			        { mouseX = (screenWidth * pos).round }
			        { mouseY = (screenHeight * pos).round };
			   }
		});
		
		
		SynthDef.writeOnce(\mouse_track, { |freq = 25|
			var clock = Impulse.kr(freq);
			SendTrig.kr(clock, [1,2], [MouseX.kr, MouseY.kr]);
			SendTrig.kr(MouseButton.kr(lag: 0), 0, 1);
		});

	}
	
	*new { |vserver|
		^super.newCopyArgs(vserver ?? Server.default).init;
	}

	init {
		var swarm;
		var index;

		// waitForBoot explicitly starts the server.
		// Later objects should use doWhenBooted.
		
		server.waitForBoot({
//			Synth.tail(server, \fx_rev_schroder_8, [ \wet, 0.6, \revtime, 2.0, \inbus, 16, \outbus, 16 ]);
			fx_bus = Bus.audio(server, 8);
			postln("fx bus: " ++ fx_bus.index);
			
			group_aux = Group.tail(RootNode(server));

			// allocate bus
			if (AtomSwarm.ambisonic)
			{
				ambi_bus = Bus.audio(server, 4);
				ambi_decoder = Synth.tail(group_aux, \atompanner_am_dec8, [ \inbus, ambi_bus.index, \outbus, fx_bus.index ]);
			};

			synth_reverb = Synth.tail(group_aux, \fx_rev_gverb_8,
			[
				\wet, 		0.2,
				\reverbtime, 	5.0,
				\damp, 		0.1,
				\inbus, 		fx_bus.index,
				\outbus, 		0,
				\amp,		1.0,
			]);
			
			"compander".postln;
//			Synth.tail(server, \fx_repl_compander, [ \inbus, 0, \outbus, 0 ]);
	
//			synth_amp = Synth.tail(server, \send_8, [ \inbus, fx_bus.index, \outbus, ambi_bus.index, \amp, 1.0 ]);
		});
		
		// swarm = AtomPlayerSwarm(1, 2, [ 1000, 800 ]).play;
		
//		swarm = AtomPlayerSwarm(1, 3, [ 1000, 800, 500 ]).play;
		swarm = AtomPlayerSwarm(1, 3, [ 800, 800, 800 ]).play;
		swarm.set = this;
		
		// multiswarm support
		swarms = [ swarm ];

//		swarms[0].activeRules = [ \brownian, \cohesion, \gravity_bounded, \sep_prop, \separation, \bound ];

		swarms[0].activeRules = [ \brownian, \cohesion, \gravity_bounded, \sep_prop, \separation, \bound ];
//		swarms[0].activeRules = [ \brownian, \cohesion, \alignment, \gravity_bounded, \sep_prop, \separation, \bound ];

		if (AtomSwarm.pheromones)
		{
			swarms[0].activeRules = swarms[0].activeRules.add(\cell);
		};

		
		swarms[0].rule_params[\sep_prop][\weight] = 0.1;
		swarms[0].rule_params[\cohesion][\weight] = 0.05;
		swarms[0].rule_params[\gravity_bounded][\weight] = 0.3;
		swarms[0].rule_params[\brownian][\weight] = 0.0;
		
		win_disp = AtomSwarmWindow(swarms[0]).play;
		
		win_ctrl = AtomCtrlWindow.new;
		
		this.loadSamples([
//			"/Users/daniel/audio/samples/speech/erotic/gasps/*.aif",
			"/Users/daniel/audio/samples/acoustic/toy piano/studio/onset/0*.aif",
			// "/Users/daniel/audio/samples/electroacoustic/penmaen piano/notes/mono/*.aif",
			// "/Users/daniel/audio/samples/speech/yorkshire_tics/*.aif",
			// "/Users/daniel/audio/samples/speech/erotic/gasps/*.aif",
//			"/Users/daniel/audio/samples/perc/sets/Roland/Roland TR-909/TR-909Kick*.wav",
//			"/Users/daniel/projects/swarm/audio/ash/*.aif",
			"/Users/daniel/projects/swarm/audio/storks/*.aif",
		]);
		
		introductions_done = [];
		
		// device 2 = sound card (with iac drivers)
		if (midi)
		{
			midicontrol = MIDIControl.new(0, 2);
			// working with faderfox relative 2's comp values
			//MIDIControl.relative = true;
			MIDIControl.relative = false;
		};
		
		win_ctrl.addControl(midicontrol, midibase + 0,
			"cohes", { |v| swarms[0].rule_params[\cohesion][\weight] = (v.value * 2.0).squared },
			"coldr", { this.createAtomOfType(\collider) }, { this.destroyAtomOfType(\collider) }
		);
		win_ctrl.addControl(midicontrol, midibase + 1,
			"gravb", { |v| swarms[0].rule_params[\gravity_bounded][\weight] = (v.value * 2.0).squared },
			"velth", { this.createAtomOfType(\velthresh) }, { this.destroyAtomOfType(\velthresh) }
		);
		win_ctrl.addControl(midicontrol, midibase + 2,
			"separ", { |v| swarms[0].rule_params[\separation][\weight] = (v.value * 2.0).squared },
			"proxi", { this.createAtomOfType(\nearest) }, { this.destroyAtomOfType(\nearest) }
		);
		win_ctrl.addControl(midicontrol, midibase + 3,
			"seppr", { |v| swarms[0].rule_params[\sep_prop][\weight] = (v.value * 40.0) },
			"depos", { this.createAtomOfType(\depositor) }, { this.destroyAtomOfType(\depositor) }
		);
		win_ctrl.addControl(midicontrol, midibase + 4,
			"bound", { |v| swarms[0].rule_params[\bound][\weight] = (v.value * 2.0).squared },
			"hornt", { this.createAtomOfType(\doppler) }, { this.destroyAtomOfType(\doppler) }
		);
		win_ctrl.addControl(midicontrol, midibase + 5,
			"brown", { |v| swarms[0].rule_params[\brownian][\weight] = (v.value * 2.0).squared },
			"gltch", { this.createAtomOfType(\glitch) }, { this.destroyAtomOfType(\glitch) }
		);
		win_ctrl.addControl(midicontrol, midibase + 6,
			"brofr", { |v| swarms[0].rule_params[\brownian][\freq] = ControlSpec(0.01, 0.5, \exp).map(v.value) },
			"chime", { this.createAtomOfType(\chimes) }, { this.destroyAtomOfType(\chimes) }
		);
		win_ctrl.addControl(midicontrol, midibase + 7,
			"inert", { |v| swarms[0].inertia = 0.9 + (v.value * 0.1) },
			"beetl", { this.createAtomOfType(\beetle) }, { this.destroyAtomOfType(\beetle) }
		);
		win_ctrl.addControl(midicontrol, midibase + 8, 
			"rate", { |v| swarms[0].rate = (v.value * 2.0).squared },
			"playr", { this.createAtomOfType(\player) }, { this.destroyAtomOfType(\player) }
		);
		win_ctrl.addControl(midicontrol, midibase + 9, 
			"revrb", { |v| synth_reverb.set(\wet, v.value * 0.5) },
			"granu", { this.createAtomOfType(\granulator) }, { this.destroyAtomOfType(\granulator) }
		);
		win_ctrl.addControl(midicontrol, midibase + 10,
			"ampli", { |v| synth_reverb.set(\amp, v.value * 2.0); },
			"effct", { this.createAtomOfType(\effector) }, { this.destroyAtomOfType(\effector) }
		);
		win_ctrl.addControl(midicontrol, midibase + 11,
			"trail", { |v|
				traillength = (v.value * 200).round;
				swarms[0].atoms.do({ |a|
					a.traillength = traillength;
				});
			}
		);
//		win_ctrl.addControl(midicontrol, midibase + 12, "trlth", { |v|
//			swarms[0].atoms.do({ |a|
//				a.trailthresh = (v.value * 50).round;
//			});
//		});


		win_ctrl.addControl(midicontrol, midibase + 12, "fbvol", { |v|
			synth_fb.set(\amp, v.value);
		});


		win_ctrl.addControl(midicontrol, midibase + 13, "fblpf", { |v| synth_fb.set(\lpf_f, \freq.asSpec.map(v.value)) });
		win_ctrl.addControl(midicontrol, midibase + 14, "feed", { |v| synth_fb.set(\wet, v.value); });

		win_ctrl.addControl(midicontrol, midibase + 15, "", { }, " OK", { this.pause });
		
		
		
		win_ctrl.view.keyDownAction = { |view, char, mods, unicode, keycode|
			var type;

			// PAUSE
			//---------------------------------------
			if (char == $ )
			{
				this.pause;
			};
			

			// SCATTER
			//---------------------------------------
			if (char == $')
			{
				this.scatter;
			};

			// UNIFY
			//----------------------------------------
			if (char == $\\)
			{
				var basenotes;
				var basenote;
				
				basenotes = swarms[0].atoms.collect(_.basenote);
				basenote = basenotes.maxItem({ |n| basenotes.occurencesOf(n) });
				
				swarms[0].atoms.do({ |atom|
					atom.basenote = basenote;
				});
			};

			// PAUSE LISTENER
			//---------------------------------------
			if (char == $.)
			{
				swarms[0].listener.paused = swarms[0].listener.paused.not;
			};

			// PAUSE ALL BUT LISTENER
			//---------------------------------------
			if (char == $,)
			{
				var atoms = swarms[0].atoms.select({ |a| a.isKindOf(AtomListener).not });
				if (atoms.size > 0)
				{
					var paused = atoms[0].paused.not;
					atoms.do({ |a| a.paused = paused });
				};
			};

			// CREATE FEEDBACK NETWORK
			//---------------------------------------
			if (char == $=)
			{
				this.createFeedNet;
			};
			if (char == $+)
			{
				this.destroyFeedNet;
			};


			// CREATE/DESTROY ATOM
			//---------------------------------------
			type = atomtypes.keys.detect({ |key| atomtypes[key][0] == char.toLower });
			if (type.notNil)
			{
				if (char.toUpper == char)
				{
					this.destroyAtomOfType(type);
				}
				{
					this.createAtomOfType(type);
				};
			};

			// QUIT
			//---------------------------------------
			if ((keycode == 53) && ((mods & 524576) > 0))
			{
				win_disp.showIntro("GAME OVER", [ 300, 300 ]);
				this.pause;
				Synth(\inst_teleport);
			};

			// STORE/RESTORE POSITION
			//---------------------------------------
			if ((keycode >= 18) && (keycode < 28))
			{
				index = keycode - 18;
				if ((mods & 131332) == 131332)
			   	{
			   		// store pos 0-9
			   		swarms[0].storeConfig(index);
			   	}
			   	{
			   		// restore pos 0-9
			   		swarms[0].restoreConfig(index);
			   		win_disp.flash;
			   	};
			};
			
			
			// postln("keyDown " ++ [ char, mods, unicode, keycode ]);
		};
		
		win_disp.view.keyDownAction = win_ctrl.view.keyDownAction;

		mouseDownActions = [
			{ |x, y|
				var winleft = win_disp.bounds.left;
				var wintop = AtomSet.screenHeight - win_disp.bounds.bottom;
				win_disp.bounds.postln;
		//		swarms[0].listener.pos[0] = x - win_disp.bounds.left;
		//		swarms[0].listener.pos[1] = y - AtomSet.screenHeight + win_disp.bounds.bottom;
			}
		];
		
		swarm.moveHook = { |self|
			if (self.atoms.size > 0)
			{
				var meanDist = self.atoms.collect({ |a| self.distances[a].values.select(_ > 0).minItem ? 0 }).sum / self.atoms.size;
				synth_reverb.set(\earlylevel, meanDist / 250 + 0.2, \taillevel, meanDist / 250 + 0.1);
			};
		};
		
		
		//---------------------------------------------------
		// Cursor tracking.
		//
		// Position and button state are only globally
		// monitored on the server side, so we must create
		// a synth to regularly relay this data to the
		// client.
		//---------------------------------------------------
		mouse_inited = true;
		mouse_responder.add;
		server.doWhenBooted({ mouse_synth = Synth(\mouse_track); });

		
		//---------------------------------------------------
		// Onset detection.
		//---------------------------------------------------
//		postln("onset resp created");
//		onset_responder = OSCresponder(nil, '/tr', { |time, rsp, msg|
//			var id = msg[2];
//			var amp = msg[3];
//			if (id == 4)
//			{
//				postln("TRIG: " ++ amp);
//				defer { this.scatter(amp * 2.0) };
//			};
//		});
//
//		onset_inited = true;
//		onset_responder.add;
//		server.doWhenBooted({
//			postln("creating scatter");
//			onset_buffer = Buffer.alloc(server, 2048, 1);
////			onset_synth = Synth(\atom_scatter, [ \bufnum, onset_buffer.bufnum ]);
//			onset_synth = Synth(\atom_scatter, [ \bufnum, 0 ]);
//		});
		CmdPeriod.add({ onset_responder.remove });
		

		//---------------------------------------------------
		// Monome 40h interface.
		// XXX TODO
		//---------------------------------------------------
		monome = Monome.new;
		monome.action = { |x, y, on|
			if (on == 1)
			{
				
			};
		};


		// 

		CmdPeriod.add({
			// do swarm cleanup
			this.destroy;
		});

	}

	destroy {
		AtomSet.mouse_inited = false;

		if (AtomSwarm.ambisonic)
		   { ambi_bus.free };

		swarms.do({ |swarm|
			swarm.destroy;
		});
		   
		fx_bus.free;
	}
	
	pause {
		if (pauserate.isNil)
		{
			pauserate = swarms[0].rate;
			swarms[0].rate = 0.0;
		}
		{
			win_disp.hideIntro;
			swarms[0].rate = pauserate;
			pauserate = nil;
		};
	}
	
	loadSamples { |paths|
		Task({
			paths.do({ |path|
				var samples = SampleLib.glob(path);
				samples.do({ |sample|
					Buffer.read(server, sample, action: {  });
					0.05.rand.wait;
				});
			});
		}).start;
	}
	
	createAtomOfType { |type, swarm|
		var atom;
		
		if (atomtypes.includesKey(type).not)
		   { postln("No such atom type: " ++ type); ^nil; };

		// postln("creating atom of type " ++ type);
		
		//--------------------------------------------------------
		// Does this set specify a custom constructor method for
		// atoms of this type? If so, use it.
		//--------------------------------------------------------
		if (atomtypes[type][3].notNil)
		{
			atom = atomtypes[type][3].value(swarms[0]);
		}
		{
			// Use default method
			atom = atomtypes[type][1].new(env: swarms[0]);
			swarms[0].add(atom);
		};
		
		atom.traillength = traillength;
		if (atom.respondsTo('outbus_'))
		{
			if (AtomSwarm.ambisonic)
			{
				atom.outbus = ambi_bus.index;
			} {
				atom.outbus = fx_bus.index
			};
		};

		
		if (introductions_done.includes(type).not)
		{
			//--------------------------------------------------------
			// Output introduction text for new atom class.
			//--------------------------------------------------------
			var title = atomtypes[type][2];
			introductions_done = introductions_done.add(type);
			win_disp.showIntro(title, atom.pos);
			this.pause;
		};
	}
	
	destroyAtomOfType { |type, swarm|
		var atoms;
		
		if (atomtypes.includesKey(type).not)
		   { postln("No such atom type: " ++ type); ^nil; };

		// destroy
		atoms = swarms[0].atoms.select({ |b| b.class == atomtypes[type][1] });
		if (atoms.notEmpty)
		   { atoms.choose.destroy };
	}
	
	outbus {
		^fx_bus.index;
	}
	
	createFeedNet
	{
		var atom, atom_last;
		
		synth_fb = Synth.head(server, \inst_fb_net, [ \outbus, fx_bus.index ]);
//		synth_fb = Synth.head(server, \inst_fb_net, [ \outbus, 0 ]);
		synth_fb.set(\wet, 0.0);
		
		atom = AtomNodeHead(\null, swarms[0]);
		atom.synth = synth_fb;
		atom.map_vel = [ \feedback,	[ 0.7, 1.5, \linear ].asSpec ];
		atom.map_axis[1] = [ \lpf_f,	[ 20, 10000, \linear ].asSpec ];
		swarms[0].add(atom);
		atom_last = atom;

		atom = AtomNode(\null, swarms[0]);
		atom.synth = synth_fb;
		atom.map_vel = [ \af_f,		[ 20, 10000, \linear ].asSpec ];
		atom.map_axis[0] = [ \af_dr,	[ 0.1, 1.0, \linear ].asSpec ];
		swarms[0].add(atom);
		atom.addPeer(atom_last);
		atom_last = atom;

		atom = AtomNode(\null, swarms[0]);
		atom.synth = synth_fb;
		atom.map_vel = [ \af_ds,	[ 0.05, 0.2, \linear ].asSpec ];
		atom.map_axis[0] = [ \hpf_f,	[ 20, 10000, \linear ].asSpec ];
		swarms[0].add(atom);
		atom.addPeer(atom_last);
		atom_last = atom;

		atom = AtomNode(\null, swarms[0]);
		atom.synth = synth_fb;
		atom.map_vel = [ \af_fr,	[ 0.7, 1.0, \linear, 0.95 ].asSpec ];
		atom.map_axis[1] = [ \af_lv,	[ 0.5, 1.0, \linear, 0.95 ].asSpec ];
		swarms[0].add(atom);
		atom.addPeer(atom_last);
		
//		if (midicontrol.controls[midibase + 13].isNil)
//		{
//			win_ctrl.addControl(midicontrol, midibase + 13, "lpf", { |v| synth_fb.set(\lpf_f, \freq.asSpec.map(v.value)) });
//			win_ctrl.addControl(midicontrol, midibase + 14, "feed", { |v| synth_fb.set(\wet, v.value); });
//		};

		win_disp.showIntro("Feedback Network", atom.pos);
		this.pause;

		^this;
	}
	
	scatter
	{ |amp = 1.0|
		swarms[0].window.flash;
		swarms[0].atoms.do({ |atom|
			atom.scatter(amp);
		});
	}
	
	destroyFeedNet
	{
		synth_fb.set(\gate, 0);
		
		swarms[0].atoms.select({ |a| a.isKindOf(AtomSynth) }).do({ |atom|
			if (atom.synth == synth_fb)
			{
				atom.destroy;
			};
		});
	}
}
