//------------------------------------------------------------------------------
// AtomSwarm
//------------------------------------------------------------------------------
// A single particle swarm. Has a fixed number of dimensions and size,
// a population of Atom objects, an array of rules which are inherited
// by these Atoms, and maintains other configuration common to its
// population.
//
// Usage:
//   var swarm = AtomSwarm.new(10, 2, [800, 600]).play;
//   swarm.activeRules = [ \cohesion, \gravity_bounded, \separation, \bound, \brownian ];
//   swarm.ruleParams[\cohesion][\weight] = 0.05;
//   swarm.ruleParams[\gravity_bounded][\weight] = 0.3;
//   swarm.storeConfig(0);
//   swarm.restoreConfig(0);
//
//   var atom = BoidMIDIChimes(swarm).midichannel_(0).midibasenote_(20);
//   swarm.add(atom);
//------------------------------------------------------------------------------

AtomSwarm {
	var <>atoms,				// Array of Atom objects 
	    <dim,					// Number of dimensions
	    <>size,				// Array containing dimension lengths [x,y,..] (pixels)
	    <>delay = 0.05,		// Delay between processing frames (s)
	    <>rate = 0.5;			// Rate at which norms are recalculated (calcs/frame)
	    						// This only really makes sense <= 1.0
	
	var <>set;				// AtomSet that I reside within
	
	var <>inertia = 0.95;
	var <>atomclass;
	
	var <>rules,
	    <>rule_params,
	    <activeRules;
	var <>attractors;  
	    
	
	// Environment norms (mean values calculated each frame)
	var <meanPos,
	    <meanVec;
	    
	var <>running = false;
	var <>limitsight = false;
	
	var <distances;
	
	// Graphical elements
	var <>color;
	var <>window;
	var <>moveHook;
	
	// Array of stored position/vector values for every Atom
	var <>config_store;
	
	// Number of output channels
	classvar <channels = 2;
	classvar <ambisonic = false;
	
	// FOR RELATIVE LISTENING
	// Width of stereo panning range in pixels.
	classvar <panrange = 150.0;
	// Maximum range at which audio can be heard.
	classvar <audiorange = 300.0;



	// Are we using pheromone structures?
	classvar <pheromones = false;

	// Pheromone parameters
	var <>cells,
	    <>cellcount = 32,
	    <>cellrange = 2,
	    <>cellaccumrate = 0.0005,
	    // <>celldecayrate = 0.0001,
	    <>celldecaymult = 0.995;
	  

	
	*new { |count = 10, dim = 3, size|
		size = size ?? Array.fill(dim, 100);
		^super.newCopyArgs(count, dim, size).init(count);
	}

	init { |count|
		rules = Atom.globalRules.copy;
		
		atomclass = Atom;
		
		// only the active rule set is calculated and applied.
		activeRules = [ \brownian, \cohesion, \separation, \bound ];

		// do not apply any global modulation to rule strengths
		rule_params = Dictionary [];
		rules.keysDo({ |ruleName|
			rule_params[ruleName] = Dictionary[ \weight -> 1.0 ];
		});
		
		// default color for atoms in this swarm
		color = Color.white;
		
		// array of stored configurations
		config_store = Array.newClear(8);
		
		atoms = Array.new(count);
		count.do({
			var rule_params = (
				cohesion: (weight: 2.0.rand),
				alignment: (weight: 2.0.rand),
				separation: (weight: 2.0.rand)
			);
			var atom = Atom(this, rule_params: rule_params);
			//atom.init;
			atoms = atoms.add(atom);
		});
		
		cells = Array.newClear(cellcount);
		cellcount.do({ |i| cells[i] = Array.fill(cellcount, 0) });
		
		attractors = Array.new(count);
	}
	
	destroy {
		atoms.do({ |atom|
			atom.destroy;
		});
	}
	
	play {
		if (running) { ^nil };
		
		running = true;
		
		TempoClock.default.sched(0, {
			this.move;
			if (running, { delay }, { nil });
		});
	}
	
	stop {
		running = false;
	}

	add { |atom|
		atoms = atoms.add(atom);
		
		^atom;
	}
	
	addDefault {
		var atom = atomclass.new(env: this);
		atoms = atoms.add(atom);
		^atom;
	}

	remove { |index|
		
	}

	
	move {
		if (limitsight.not)
		   { this.calculateNorms; };
		
		distances = Dictionary[];
		atoms.do({ |atom|
			distances[atom] = Dictionary[];
			atoms.do({ |peer|
				distances[atom][peer] = sqrt((atom.pos - peer.pos).squared.sum);
			});
		});
		
		atoms.do({ |atom|
			var cellx, celly;
			
			if (limitsight)
			   { atom.calculateNorms; };
			atom.move(this);
			
			
			// Calculate atom's pheromone grid reference, and increase
			// pheromone quantity in this cell.
			cellx = floor(atom.pos[0] / (size[0] / cellcount));
			celly = floor(atom.pos[1] / (size[1] / cellcount));
			cells[celly][cellx] = cells[celly][cellx] + cellaccumrate;
		});
		
//		cells = (cells - celldecayrate).clip(0, 1);

		// Apply multiplicative decay of pheromone (following Bonabeau et
		// al), and clip within bounds.
		cells = (cells * celldecaymult).clip(0, 1);

		attractors.do({ |attractor|
			attractor.move;
		});
		
		if (moveHook.notNil)
		   { moveHook.value(this) };
	}
	
	
	//----------------------------------------------------------------
	// Calculate mean position and velocity for population, then
	// used for rules such as cohesion. Avoids recalculating
	// global norms for every boid in the swarm.
	//----------------------------------------------------------------
	
	calculateNorms {
		var pos = 0, vec = 0;
		var attrstrength = 0;
		
		atoms.do({ |atom|
			pos = pos + atom.pos;
			vec = vec + atom.vec;
		});
		
		attractors.do({ |attr|
			pos = pos + (attr.pos * attr.effectivestrength);
			attrstrength = attrstrength + attr.effectivestrength;
		});

		meanPos = pos / (atoms.size + attrstrength);
		meanVec = vec / (atoms.size + attrstrength);
	}

	meanVelocity {
		^(atoms.collect(_.velocity).sum / atoms.size);
	}

	addAttractor { |pos, strength = 1.0|
		var attr = Attractor(this, pos, strength);
		attractors = attractors.add(attr);
	}
	
	
	//----------------------------------------------------------------
	// Assign new set of active rules. These are then propagated
	// to the entire population. (Should really be inherited
	// dynamically...)
	//----------------------------------------------------------------
	
	activeRules_ { |newActiveRules|
		activeRules = newActiveRules;
		atoms.do({ |atom|
			atom.activeRules = activeRules;
		});
	}



	//----------------------------------------------------------------
	// Store the position and velocity of every Atom into a given
	// storage index, available to be restored later.
	//----------------------------------------------------------------
	
	storeConfig { |n = 0|
		config_store[n] = Dictionary[];
		// XXX: also need to store vector!
		atoms.do({ |atom|
			config_store[n][atom] = [ atom.pos.copy, atom.vec.copy ];
		});
	}
	
	restoreConfig { |n = 0|
		if (config_store[n].isNil)
		   { ^nil; };
		   
		atoms.do({ |atom|
			// atom may have been created since config was saved
			if (config_store[n].includesKey(atom))
			{
				atom.pos = config_store[n][atom][0].copy;
				atom.vec = config_store[n][atom][1].copy;
			};
		});
		thisThread.randSeed = 1.5;
	}
}



AtomGranulatorSwarm : AtomSwarm {
	init {
		super.init;
		atomclass = AtomGranulator;
	}
	
	addWithSample { |sample, rule_params|
		var atom = AtomGranulator(sample, this, rule_params: rule_params);
		atoms = atoms.add(atom);
		^atom;
	}
}









//
//
//


AtomPlayerSwarm : AtomSwarm {
	var <>audiorange;
	var <>listener;
	
	init {
		super.init;
		listener = AtomListener(this);
		listener.init;
		
		audiorange = 180;
		
		atoms = [ listener ];
	}
	
	addWithSample { |sample, rule_params|
		var atom = AtomPlayer(sample, this, rule_params: rule_params);
		atom.init;
		atoms = atoms.add(atom);
		^atom;
	}
}

