//--------------------------------------------------------------------
// AtomWindow
//--------------------------------------------------------------------
// Window superclass, implements generic framing and UI methods for
// unified appearance.
//
// Usage:
//   var win = AtomWindow.new(title, Rect(xpos, ypos, xsize, ysize))
//
//--------------------------------------------------------------------

AtomWindow : SCWindow
{
	var <>title,
	    <>bounds;

	//------------------------------------------------------
	// Graphical elements
	//------------------------------------------------------
	var <>gBorder,
	    <>gPanel;
	var <>gTop,
	    <>gTopText;
	
	*new { |vtitle, vbounds|
		^super.new(vtitle, vbounds).title_(vtitle).bounds_(vbounds).init;
	}
	
	init {
		view.background = Color(0, 0, 0);
		
		gBorder = SCCompositeView(this, Rect(1, 20, bounds.width - 2, bounds.height - 21));
		gBorder.background = Color(0.4, 0.4, 0.4);
		gBorder.background = Color(0.15, 0.17, 0.1);

		gPanel = SCCompositeView(gBorder, RectIn(gBorder, 5, 5, bounds.width - 12, bounds.height - 31));
		gPanel.background = Color(0.95, 0.96, 0.90);
		
		gPanel.background = Color(0.1, 0.1, 0.1);
		
		// Higher contrast is more suitable for video projector
		//gPanel.background = Color(0.02, 0.02, 0.02);

		
		gTop = SCCompositeView(gBorder, Rect(1, 1, bounds.width - 2, 18));
		gTop.background = Color(0.7, 0.75, 0.5);
		
		gTopText = SCStaticText(gTop, RectIn(gTop, 4, 0, 200, 16));
		gTopText.font = Font("Rockwell", 12);
		gTopText.stringColor = Color.white;
		gTopText.string = title;

		this.front;
		
		// Close windows when server processing is halted
		CmdPeriod.add({ this.close; });
	}
}


//--------------------------------------------------------------------
// AtomCtrlWindow
//--------------------------------------------------------------------
// Contains a series of control elements, used to affect swarm
// parameters during performance.
//--------------------------------------------------------------------

AtomCtrlWindow : AtomWindow {
	var <>controlFuncs, <>controlKnobs;
	
	*new {
		^super.new('AtomControls', Rect(50, 800, 210, 320));
	}
	
	addControl { |midicontrol, number, string, fn, b_string, b_fn, b_fn_d|
		var knob, label;
		var xoff, yoff;
		
		xoff = 40 * (controlFuncs.size % 4);
		yoff = 65 * (div(controlFuncs.size, 4));
		
		knob = Knob(gPanel, Rect(30 + xoff, 40 + yoff, 30, 30));
		knob.color[0] = Color(0.8, 0.8, 0.7);
		knob.color[2] = Color(0.3, 0.3, 0.3);
		knob.action = fn;
		knob.canFocus = false;
		
		label = SCStaticText(gPanel, Rect(30 + xoff, 70 + yoff, 32, 20));
		label.stringColor = Color(0.9, 0.9, 0.9);
		label.font = Font("Silkscreen", 8);
		label.string = string;

		// now using relative
		if (midicontrol.notNil)
		{
//			midicontrol.addControl(number, { |v| { knob.value = v }.defer; fn.value(v) });
			midicontrol.addControl(number, { |v| [ number, v ].postln; { knob.value = v }.defer; fn.value(v) });
//			midicontrol.addControl(number, { |v| [ number, v ].postln; { knob.value = knob.value + v }.defer; fn.value(v) });
		};
		
		controlKnobs = controlKnobs.add(knob);
		controlFuncs = controlFuncs.add(fn);
		
		//------------------------------------------------------
		// Button functions, to support faderfox's push-pots
		//------------------------------------------------------
		
		if (b_string.notNil)
		{
			label = SCStaticText(gPanel, Rect(30 + xoff, 80 + yoff, 32, 20));
			label.stringColor = Color(0.8, 0.4, 0.3);
			label.font = Font("Silkscreen", 8);
			label.string = b_string;
			
			if (midicontrol.notNil)
			{
				midicontrol.addNoteOn(number, { |on| if (on > 0) { defer { b_fn.value } } });
				midicontrol.addNoteOn(number + 16, { |on| if (on > 0) { defer { b_fn_d.value } } });
			};
		};
	}

}




//--------------------------------------------------------------------
// AtomSwarmWindow
//--------------------------------------------------------------------
// Displays the particle swarm environment, updated in real time.
//--------------------------------------------------------------------


AtomSwarmWindow : AtomWindow {
	var <>swarms;
	var <>gDark, <>gDarkOn = false;
	var <>running = false;
	var <>flashing = false;
	
	var <>controlFuncs, <>controlKnobs;
	
	var <>gIntroText,
	    <>gIntroCircle;
	
	*new { |swarm|
		^super.new('AtomSwarm', Rect(50, 50, swarm.size[0], swarm.size[1])).addSwarm(swarm);
	}
	
	init {
		super.init;
		controlFuncs = [];
		controlKnobs = [];
		gTop.background = Color(0.3, 0.3, 0.3);
	}
	
	addSwarm { |swarm|
		swarms = swarms.add(swarm);
		swarm.window = this;
	}
	
	addControl { |midicontrol, number, string, fn|
		var knob, label;
		
		knob = Knob(gPanel, Rect(30 + (40 * controlFuncs.size), 40, 30, 30));
		knob.color[0] = Color(0.8, 0.8, 0.7);
		knob.color[2] = Color(0.3, 0.3, 0.3);
		knob.action = fn;
		knob.canFocus = false;
		
		label = SCStaticText(gPanel, Rect(30 + (40 * controlFuncs.size), 70, 30, 20));
		label.stringColor = Color(0.9, 0.9, 0.9);
		label.font = Font("Silkscreen", 8);
		label.string = string;
		
		midicontrol.addControl(number, { |v| { knob.value = v }.defer; fn.value(v) });
		midicontrol.addControl(number + 16, { |v| { knob.value = v }.defer; fn.value(v) });
		
		controlKnobs = controlKnobs.add(knob);
		controlFuncs = controlFuncs.add(fn);
	}
	
	play {
		if (running) { ^nil };
		
		running = true;

		gPanel.mouseDownAction = { |self, x, y|
			if (swarms[0].respondsTo(\recorder))
			   { swarms[0].recorder.pos = [ x, y ] };
		};
		
		drawHook = { this.draw };
		
		// can only catch keydown actions if the window has a border...
		view.keyDownAction = { |view, char, mods, unicode, keycode|
			postln("keyDown " ++ [ char, mods, unicode, keycode ]);
		};

		gIntroText = SCStaticText(gPanel, RectIn(gPanel, 40, 40, 400, 86));
		gIntroText.font = Font("Silkscreen", 44);
		gIntroText.stringColor = Color(1, 1, 1, 0.8);
		gIntroText.string = "";

		gDark = SCWindow(\dark, Rect(0, 0, SCWindow.screenBounds.width, SCWindow.screenBounds.height), border: false);
//		gDark.view.background = Color(0.4, 0.45, 0.55);
		gDark.view.background = Color(0.05, 0.05, 0.15);
		gDark.alpha = 0.95;
		// don't open darkening view - this makes it impossible to execute code!
		if (gDarkOn) { gDark.front; this.front; "gDark".postln; };
		
		CmdPeriod.add({ gDark.close; });
		
		AppClock.sched(0, {
			this.refresh;
			if (running, { swarms[0].delay }, { nil });
		});
			
	}

	draw {
		if (flashing)
		{
			//gPanel.background = Color(0.2, 0.2, 0.2);
			
			// Better for projector
			gPanel.background = Color(0.1, 0.1, 0.1);
			flashing = false;
		};
		
		//--------------------------------------------------------------------
		// If pheromones are active, draw a grid with suitable X/Y dimensions
		// and colour each cell accordingly - higher pheromone concenration
		// -> higher luminosity.
		//--------------------------------------------------------------------
		if (AtomSwarm.pheromones)
		{
			swarms[0].cellcount.do({ |y|
				var barwidth = bounds.width / swarms[0].cellcount;
				
	 			swarms[0].cellcount.do({ |x|
					var alpha = swarms[0].cells[y][x].clip(0, 1);
					var rect = Rect(x * barwidth, y * barwidth, barwidth, barwidth);
					
					if (alpha > 0)
					{
						Color(1, 1, 1, swarms[0].cells[y][x].clip(0, 1)).set;
						Pen.fillRect(rect);
					};
					
					Color(0, 0, 0, 0.1).set;
					Pen.strokeRect(rect);
				});
			});
		}
		{
			12.do({ |n|
				var xpos = (n + 1) * (bounds.width / 13.0);
				Pen.moveTo(Point(xpos, 0));
				Pen.lineTo(Point(xpos, bounds.height));
			});
	
			12.do({ |n|
				var ypos = (n + 1) * (bounds.height / 13.0);
				Pen.moveTo(Point(0, ypos));
				Pen.lineTo(Point(bounds.width, ypos));
			});
		};
		
		Color(0.0, 0.0, 0.0, 1.0).set;
		Pen.stroke;

		Pen.use
		{
			swarms.do({ |swarm|
				swarm.atoms.do({ |atom|
					atom.draw;
				});
				
				swarm.attractors.do({ |attr|
					attr.draw;
				});
			});

		};

		if (gIntroCircle.notNil)
		{
			Pen.use {
				Color(1, 1, 1, 0.5).set;
				Pen.width = 0;
				Pen.strokeOval(gIntroCircle);
			};
		};
	}
	
	flash {
		gPanel.background = Color(0.15, 0.15, 0.15);
		flashing = true;
	}


	//------------------------------------------------------
	// Display text to introduce a new atom class
	//------------------------------------------------------
	showIntro { |string, pos|
		var left, top;
		
		left = pos[0];
		top = pos[1] - 30;
		
		postln("left " ++ left ++ ", top " ++ top);
		if (left > 500)
		   { left = left - 300 };
		
		if (top > 500)
		   { top = top - 100; };
		gIntroText.bounds = Rect(left, top, 400, 150);
		gIntroText.string = string.toUpper;
		gIntroText.refresh;
		gIntroCircle = Rect(pos[0] - 15, pos[1] - 15, 30, 30);
	}

	hideIntro {
		gIntroText.string = "";
		gIntroCircle = nil;
	}
}


